﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading;

using DevExpress.Map;
using DevExpress.XtraMap;

namespace TestProject
{
    /// <summary>
    /// 치료 클러스터러
    /// </summary>
    public class CureClusterer : IClusterer
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Field
        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region Field

        /// <summary>
        /// 분주함 여부
        /// </summary>
        private bool isBusy;

        /// <summary>
        /// 클러스터 카운트
        /// </summary>
        private int clusterCount = 10;

        /// <summary>
        /// 맵 항목 컬렉션
        /// </summary>
        private MapItemCollection mapItemCollection;

        /// <summary>
        /// 맵 데이터 어댑터
        /// </summary>
        private IMapDataAdapter mapDataAdapter;

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Property
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 분주함 여부 - IsBusy

        /// <summary>
        /// 분주함 여부
        /// </summary>
        public bool IsBusy
        {
            get
            {
                return this.isBusy;
            }
        }

        #endregion
        #region 항목 컬렉션 - Items

        /// <summary>
        /// 항목 컬렉션
        /// </summary>
        public MapItemCollection Items
        {
            get
            {
                return this.mapItemCollection;
            }
        }

        #endregion
        #region 클러스터 카운트 - ClusterCount

        /// <summary>
        /// 클러스터 카운트
        /// </summary>
        public int ClusterCount
        {
            get
            {
                return this.clusterCount;
            }
            set
            {
                if(value < 1)
                {
                    throw new ArgumentOutOfRangeException("The ClusterCount should be larger than 1.");
                }

                this.clusterCount = value;
            }
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 생성자 - CureClusterer()

        /// <summary>
        /// 생성자
        /// </summary>
        public CureClusterer()
        {
            this.isBusy = false;
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 클러스터 만들기 - Clusterize(sourceEnumerable, viewport, sourceChanged)

        /// <summary>
        /// 클러스터 만들기
        /// </summary>
        /// <param name="sourceEnumerable">소스 열거 가능 목록</param>
        /// <param name="viewport">뷰포트</param>
        /// <param name="sourceChanged">소스 변경 여부</param>
        public void Clusterize(IEnumerable<MapItem> sourceEnumerable, MapViewport viewport, bool sourceChanged)
        {
            Thread thread = new Thread
            (
                () => {

                    this.isBusy = true;

                    if(sourceChanged)
                    {
                        this.mapItemCollection = GetMapItemCollection(sourceEnumerable);

                        this.mapDataAdapter.OnClustered();
                    }

                    this.isBusy = false;
                }
            );

            thread.Start();
        }

        #endregion
        #region 소유자 설정하기 - SetOwner(mapDataAdapter)

        /// <summary>
        /// 소유자 설정하기
        /// </summary>
        /// <param name="mapDataAdapter">맵 데이터 어댑터</param>
        public void SetOwner(IMapDataAdapter mapDataAdapter)
        {
            this.mapDataAdapter = mapDataAdapter;
        }

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region 가장 가깝게 배열하기 - AssignClosest(cluster, sourceList)

        /// <summary>
        /// 가장 가깝게 배열하기
        /// </summary>
        /// <param name="cluster">클러스터</param>
        /// <param name="sourceList">소스 리스트</param>
        private void AssignClosest(Cluster cluster, List<Cluster> sourceList)
        {
            if(sourceList.Count < 2)
            {
                throw new ArgumentOutOfRangeException("Clusters count should be larger than 2.");
            }

            Cluster distancableCluster = sourceList[0];

            if(distancableCluster == cluster)
            {
                distancableCluster = sourceList[1];
            }

            cluster.ClosestCluster = distancableCluster;

            for(int i = 0; i < sourceList.Count; ++i)
            {
                distancableCluster = sourceList[i];

                if(distancableCluster == cluster)
                {
                    continue;
                }

                double distance = cluster.GetDistance(distancableCluster);

                if(distance < cluster.DistanceToClosest)
                {
                    cluster.ClosestCluster = distancableCluster;
                }
            }
        }

        #endregion
        #region 삽입하기 - Insert(targetList, cluster)

        /// <summary>
        /// 삽입하기
        /// </summary>
        /// <param name="targetList">타겟 리스트</param>
        /// <param name="cluster">클러스터</param>
        private void Insert(List<Cluster> targetList, Cluster cluster)
        {
            for(int i = 0; i < targetList.Count; i++)
            {
                if(targetList[i].DistanceToClosest > cluster.DistanceToClosest)
                {
                    targetList.Insert(i, cluster);

                    return;
                }
            }

            targetList.Add(cluster);
        }

        #endregion
        #region 클러스터 리스트 배열하기 - ArrangeClusterList(sourceList)

        /// <summary>
        /// 클러스터 리스트 배열하기
        /// </summary>
        /// <param name="sourceList">소스 리스트</param>
        /// <returns>클러스터 리스트</returns>
        private List<Cluster> ArrangeClusterList(List<Cluster> sourceList)
        {
            List<Cluster> targetList = new List<Cluster>();

            for(int i = 0; i < sourceList.Count; ++i)
            {
                Cluster cluster = sourceList[i];

                AssignClosest(cluster, sourceList);

                Insert(targetList, cluster);
            }

            return targetList;
        }

        #endregion
        #region 가장 가까운 클러스터 병합하기 - MergeCloserstCluster(targetList)

        /// <summary>
        /// 가장 가까운 클러스터 병합하기
        /// </summary>
        /// <param name="targetList">타겟 리스트</param>
        private void MergeCloserstCluster(ref List<Cluster> targetList)
        {
            if(targetList.Count < 2)
            {
                throw new ArgumentOutOfRangeException("Clusters count should be larger than 2.");
            }

            Cluster cluster1 = targetList[0];
            Cluster cluster2 = cluster1.ClosestCluster;

            targetList.RemoveAt(0);

            targetList.Remove(cluster2);

            Cluster newCluster = Cluster.Merge(cluster1, cluster2);

            targetList.Add(newCluster);

            targetList = ArrangeClusterList(targetList);
        }

        #endregion
        #region 맵 항목 컬렉션 구하기 - GetMapItemCollection(sourceEnumerable)

        /// <summary>
        /// 맵 항목 컬렉션 구하기
        /// </summary>
        /// <param name="sourceEnumerable">소스 열거 가능 목록</param>
        /// <returns>맵 항목 컬렉션</returns>
        private MapItemCollection GetMapItemCollection(IEnumerable<MapItem> sourceEnumerable)
        {
            List<MapItem> mapItemList = new List<MapItem>();

            List<Cluster> clusterList = new List<Cluster>();

            foreach(MapItem mapItem in sourceEnumerable)
            {
                ISupportCoordLocation supportCoordLocation = mapItem as ISupportCoordLocation;

                if(supportCoordLocation != null)
                {
                    clusterList.Add(Cluster.Initialize(supportCoordLocation));
                }
                else
                {
                    mapItemList.Add(mapItem);
                }
            }

            clusterList = ArrangeClusterList(clusterList);

            while(clusterList.Count > ClusterCount)
            {
                MergeCloserstCluster(ref clusterList);
            }

            MapItemCollection mapItemCollection = new MapItemCollection(mapDataAdapter);

            for(int i = 0; i < clusterList.Count; i++)
            {
                Cluster cluster = clusterList[i];

                MapDot mapDot = new MapDot()
                {
                    Location = new GeoPoint(cluster.CenterPoint.Y, cluster.CenterPoint.X),
                    Size     = 100
                };

                mapDot.ClusteredItems       = cluster.SupportCoordLocationList.Select(item => item as MapItem).ToList();
                mapDot.TitleOptions.Pattern = mapDot.ClusteredItems.Count.ToString();

                mapItemCollection.Add(mapDot);
            }

            mapItemCollection.AddRange(mapItemList);

            return mapItemCollection;
        }

        #endregion
    }
}